import { rgbIndices } from '../constants.js';

// an ouput module intended to provide a preview of the data being gathered.
// it performs a binary search to determine the color at a given pixel.
// the pattern can tell it to display blocks of any size, so it can be used
// to display a preview in increasing resolution.
export default function preview(canvas) {
  // r, g, b
  const channelCount = 3;
  
  const ctx = canvas.getContext('2d');
  ctx.globalCompositeOperation = 'copy';
  
  // a Uint8Array containing the actual color values.
  let bytes;
  // a Uint8Array containing the next bit to be tested in each byte of `bytes`.
  let nextZs;
  
  // update `bytes` and `nextZs` to reflect the result of a threshold test,
  // but don't update the canvas.
  function set({ x, y, z, channel }, active) {
    const { width, height } = canvas;
    
    if(!(channel in rgbIndices)) {
      // update all of r, g, and b at once for greyscale.
      for(const channel in rgbIndices) {
        set({ x, y, z, channel }, active);
      }
      return;
    }
    
    const index = (x + y * width) * channelCount + rgbIndices[channel];
    if(active) {
      bytes[index] |= nextZs[index];
    }
    nextZs[index] >>= 1;
  }
  
  return {
    // this must be called before using the output. it may be called as many
    // times as necessary.
    reset(width, height) {
      canvas.width = width;
      canvas.height = height;
      ctx.fillStyle = '#000';
      ctx.fillRect(0, 0, width, height);
      bytes = new Uint8Array(width * height * channelCount);
      nextZs = new Uint8Array(width * height * channelCount);
      // set each byte in nextZs to the binary value '10000000', or 128.
      for(let i = 0; i < width * height * channelCount; ++i) {
        nextZs[i] = 0x80;
      }
    },
    // return the next z-value to be tested for a given pixel and channel.
    nextZ({ x, y, channel }) {
      const { width } = canvas;
      const index = (x + y * width) * channelCount + (rgbIndices[channel] |0);
      return (bytes[index] | nextZs[index]) / 255;
    },
    // update the preview to reflect the result of a threshold test.
    put({ x, y, z, channel }, active, blockWidth, blockHeight) {
      if(blockWidth === undefined) {
        blockWidth = 1;
      }
      if(blockHeight === undefined) {
        blockHeight = blockWidth;
      }
      
      const { width, height } = canvas;
      set({ x, y, z, channel }, active);
      
      const baseIndex = (x + y * width) * channelCount;
      ctx.fillStyle = `rgb(
                           ${bytes[baseIndex + 0]},
                           ${bytes[baseIndex + 1]},
                           ${bytes[baseIndex + 2]}
                          )`;
      ctx.fillRect(x, y, blockWidth, blockHeight);
    },
    finish() {
      
    },
  };
}
