import { requestIdleCallback, cancelIdleCallback } from '../idle.js';
import { rgbIndices } from '../constants.js';

// `sameOrigin`, so called because it only works on same-origin content.
// this one doesn't use any exploits, so it should always work. you can use it
// to check the baseline speed of my infrastructure and to make sure things
// are working properly.
export default function sameOrigin(img, pattern) {
  const { width, height } = img;
  
  // we need to draw the image to a canvas before we can read its pixels.
  const canvas = document.createElement('canvas');
  canvas.width = width;
  canvas.height = height;
  const ctx = canvas.getContext('2d');
  ctx.drawImage(img, 0, 0);
  // and here they are. 1337 h4x complete. time to pack up, boys.
  const imageData = ctx.getImageData(0, 0, width, height);
  
  // returns true if the channel `channel` of the pixel at (x, y) has a value
  // greater than or equal to `z` when normalized to the range [0..1].
  // these are the only sorts of queries we can perform on cross-origin images,
  // so we're emulating that restriction here.
  function measure({ x, y, z, channel }) {
    const index = (x + y * imageData.width) * 4;
    
    let channelValue;
    if(channel in rgbIndices) {
      channelValue = imageData.data[index + rgbIndices[channel]] / 255;
    } else {
      // if it isn't r, g, or b, assume it's brightness.
      // yes, it's sloppy, but this is just a PoC.
      
      const r = imageData.data[index + 0] / 255,
            g = imageData.data[index + 1] / 255,
            b = imageData.data[index + 2] / 255;
      
      // http://alienryderflex.com/hsp.html
      channelValue = Math.sqrt(0.299*r*r + 0.587*g*g + 0.114*b*b);
    }
    
    return channelValue >= z;
  }
  
  // since we're passing data back into the iterator, the first `next()` has
  // to be a special case. we don't have any data yet, after all.
  let next = pattern.next();
  // storing the callback ID will let us cancel the job later.
  let callbackID = requestIdleCallback(function step(deadline) {
    try {
      do {
        if(next.done) {
          // the pattern has ended, and the job is complete.
          return;
        }
        // measure the point, then pass the result back, then get the next point
        // to measure.
        next = pattern.next(measure(next.value));
      } while(deadline.timeRemaining() > 0);
      
      // there's still more work to do, so request another idle callback.
      callbackID = requestIdleCallback(step);
    } catch(e) {
      pattern.throw(e);
    }
  });
  
  return {
    cancel() {
      cancelIdleCallback(callbackID);
    },
  };
};
